from datetime import datetime, date, timedelta, timezone
from itertools import product
import numpy as np
import psycopg2
from numba import jit, prange
from time import sleep
import functools
import os
import re
from openpyxl.utils import get_column_letter
from openpyxl import Workbook
from openpyxl.styles import Font
from openpyxl.chart import LineChart, Reference
from openpyxl.chart.label import DataLabelList
from openpyxl.drawing.text import Paragraph, ParagraphProperties, CharacterProperties
from openpyxl.chart.text import RichText


FULL_FILENAME_XLSX = r'/home/ext-balyasnikov.d@main.russianpost.ru/dion-stat/report.xlsx'      # полный путь и название файла, "r" перед кавычками (r'...') НЕ ТРОГАТЬ!
DEPTH_DAYS_QTY = 111         # глубина запроса, дней
# креды для подключения к БД:
USER_NAME = 'postgres'              # login
PASSWD = 'sdfgsdfgsdfg'               # password
HOST = '10.193.79.204'              # "10.10.10.1" -- ip БД
PORT = '5432'               # порт
DB_NAME = 'backend'       # название БД


def pg_replica_connection():
    """Подключает к БД

    :return connection:
    """
    connection = psycopg2.connect(
        user=USER_NAME,
        password=PASSWD,
        host=HOST,
        port=PORT,
        database=DB_NAME
    )
    return connection


def db_retry_decorator(func):
    @functools.wraps(func)
    def wrapper(db_data, date_from, date_to, cursor, connection, *args, **kwargs):
        max_attempts = 3        # максимум попыток
        timeout_sec = 3         # длительность таймаута в сек
        last_exception = None   # ошибки будут храниться тут
        #
        for i in range(1, max_attempts + 1):
            try:
                # Выполняем основную функцию
                return func(db_data, date_from, date_to, cursor, connection, *args, **kwargs)
            except psycopg2.Error as e:
                last_exception = e
                print(f"Попытка {i} завершилась ошибкой: {e}")
                print(f"Повторное подключение через {timeout_sec} сек...")
                # Пытаемся закрыть текущее соединение
                try:
                    if cursor and not cursor.closed:
                        cursor.close()
                except Exception as e:
                    print(f"Ошибка при закрытии курсора соединения: {e}")
                try:
                    if connection and not connection.closed:
                        connection.close()
                except Exception as e:
                    print(f"Ошибка при закрытии соединения: {e}")
                #
                sleep(timeout_sec)
                # Восстанавливаем подключение
                try:
                    connection = pg_replica_connection()
                    cursor = connection.cursor()
                except Exception as e:
                    last_exception = e
                    print(f"Ошибка при восстановлении соединения: {e}")
                    # Если восстановление не удалось, то следующая попытка тоже будет с None
        # Если ни одна попытка не удалась, пробрасываем последнее исключение
        raise last_exception
    return wrapper


def process_db(user_sessions_list: list, date_from: date, date_to: date):
    """Собирает timestamp начала и конца юзерсессий по дименшенам и дням.
    Также возвращает минимальную и максимальную дату-время в выгрузке

    :param user_sessions_list: [[us_created_at, us_ended_at, event_id, user_id, slug], ...]
    :type user_sessions_list: list[list]
    :param date_from: дата начала интервала
    :type date_from: date
    :param date_to: дата окончания периода
    :type date_to: date
    :return grouped_session_dic: {(org_id,): [[ts_created_at, ts_ended_at], ...], ...} (**dict**)
    :return min_dt: минимальная дата-время в выгрузке (**datetime**)
    :return max_dt: максимальная дата-время в выгрузке (**datetime**)
    """
    grouped_session_dic = {}
    dt_now_msk = datetime.now(timezone.utc).replace(tzinfo=None) + timedelta(hours=3)
    dt_from = datetime.fromisoformat(f'{date_from}')
    dt_to = datetime.fromisoformat(f'{date_to}')
    dt_to = dt_now_msk if dt_to > dt_now_msk else dt_to
    min_dt = dt_from
    max_dt: datetime = user_sessions_list[0][1].replace(microsecond=0) if user_sessions_list else dt_to
    for row in user_sessions_list:
        created_at, ended_at, event_id, user_id, slug = row
        org_id = None
        created_at: datetime = created_at.replace(microsecond=0)
        created_at = min_dt if created_at < min_dt else created_at
        # добавляем 1 мин к началу сессии, если сессия не закрыта:
        ended_at: datetime = ended_at.replace(microsecond=0) if ended_at else created_at + timedelta(minutes=1)
        # излекаем максимальную datetime
        ended_at = ended_at if ended_at < dt_to else dt_to - timedelta(seconds=1)
        max_dt = ended_at if ended_at > max_dt else max_dt
        # получаем целочисленные timestamp:
        ts_created_at = int(created_at.timestamp())
        ts_ended_at = int(ended_at.timestamp())
        # собираем словарь с сессиями
        total_val = None
        lists = [[d, ] for d in (org_id,)]
        kkey_combinations = list(product(*lists))
        for kkey_tuple in kkey_combinations:
            session_list: list = grouped_session_dic.get(kkey_tuple, [])
            session_list.append([ts_created_at, ts_ended_at])
            grouped_session_dic[kkey_tuple] = session_list
    return grouped_session_dic, min_dt, max_dt


def divide_period_per_second(min_dt: datetime, max_dt: datetime):
    """Разбивает промежуток времени между **min_dt** и **max_dt** по секундам.
    Возвращает посекундный массив с timestamp и словарь с соответствующими индексами для этого массива для каждого агрегирующего промежутка времени **dt_val**

    :param min_dt: минимальная дата-время в выгрузке
    :type min_dt: datetime
    :param max_dt: максимальная дата-время в выгрузке
    :type max_dt: datetime
    :return ts_sec_array: массив секунд [min_ts, ..., max_ts] (**ndarray**)
    :return ts_sec_idx_dic: индексы для каждого dt_val в массиве секунд {dt_val: [min_i, max_i], ...} (**dict**)
    """
    dtime_sec_list = []
    while min_dt <= max_dt:
        dtime_sec_list.append(min_dt)
        min_dt += timedelta(seconds=1)
    dtime_sec_array: list[datetime] = np.array(dtime_sec_list)
    ts_sec_list = []
    ts_sec_idx_dic = {}
    for i, dt in enumerate(dtime_sec_array):
        ts_sec = int(dt.timestamp())
        ts_sec_list.append(ts_sec)
        dt_val = dt.date()
        min_i, max_i = ts_sec_idx_dic.get(dt_val, [i, i])
        if i > max_i:
            max_i = i
        ts_sec_idx_dic[dt_val] = [min_i, max_i]
    ts_sec_array = np.array(ts_sec_list)
    return ts_sec_array, ts_sec_idx_dic


@jit
def calc_overlaps_qty(ts_sec_array: np.ndarray, ts_intrvls_array: np.ndarray):
    """
    Вычисляет для каждой секунды из ts_sec_array кол-во одновременных интервалов в ts_intrvls_array.
    Возвращает массив длиной ts_sec_array

    :param ts_sec_array: массив секунд [min_ts, ..., max_ts] (ndarray)
    :param ts_intrvls_array: массив интервалов для которых считаем одновременность [ [ts_start, ts_end], ... ] (ndarray)
    :return: массив кол-ва пересечений для каждой секунды [conc_qty 0, ... , conc_qty N] (ndarray)
    """
    overlaps_ts_len = len(ts_sec_array)
    ts_intrvls_len = len(ts_intrvls_array)
    ts_array = np.full(overlaps_ts_len, 0)
    ts_min = ts_sec_array[0]
    for j in prange(ts_intrvls_len):
        ts_intrvl_start, ts_intrvl_end = ts_intrvls_array[j]
        while ts_intrvl_start <= ts_intrvl_end:
            dts_idx = ts_intrvl_start - ts_min
            ts_array[dts_idx] += 1
            ts_intrvl_start += 1
    return ts_array


def calculate_max_concurrent_for_db(db_maxload: list, grouped_session_dic: dict, ts_sec_array: np.ndarray, ts_sec_idx_dic: dict):
    """Считает максимум одновременных для каждой группы дименшенов за агрегирующий промежуток времени **dt_val**.
    Возвращает таблицу для записи в БД

    :param db_maxload: [[dt_val, max_val], ...]
    :type db_maxload: list[list]
    :param grouped_session_dic: {(org_id,): [[ts_created_at, ts_ended_at], ...], ...}
    :type grouped_session_dic: dict
    :param ts_sec_array: массив секунд [min_ts, ..., max_ts-1]
    :type ts_sec_array: ndarray
    :param ts_sec_idx_dic: индексы для каждого dt_val в массиве секунд {dt_val: [min_i, max_i], ...}
    :type ts_sec_idx_dic: dict
    :return db_maxload: [[dt_val, max_val], ...] (**list[list]**)
    """
    for kkey_tuple, session_list in grouped_session_dic.items():
        org_id, = kkey_tuple
        ts_session_array = np.array(session_list)
        ts_array: np.ndarray = calc_overlaps_qty(ts_sec_array, ts_session_array)
        for dt_val, min_max_i in ts_sec_idx_dic.items():
            min_i, max_i = min_max_i
            max_val = ts_array[min_i:max_i+1].max()
            db_maxload.append([dt_val, int(max_val)])
    return db_maxload


@db_retry_decorator
def get_all_confs_data(all_confs_list: list, date_from: date, date_to: date, cursor=None, connection=None):
    """Выгружает ивент сессии (конфы) с дименшенами за период, начиная с **date_from** и до **date_to**

    :return all_confs_list: [[es_id, event_id, es_created_at, es_ended_at, slug], ...]
    :type all_confs_list: list[list]
    :param date_from: дата начала периода
    :type date_from: date
    :param date_to: дата конца периода
    :type date_to: date
    :param cursor: объект курсора подключения к БД
    :type cursor: cursor | None
    :param connection: объект подключения к БД
    :type connection: connection | None
    :return all_confs_list: [[es_id, event_id, es_created_at, es_ended_at, slug], ...] (**list[list]**)
    :return cursor: объект курсора подключения к БД (**cursor** | **None**)
    :return connection: объект подключения к БД (**connection** | **None**)
    """
    cursor.execute(f"""
    select  es.id as es_id,
            es.event_id,
            es.created_at + interval '3 H' as es_created_at,
            es.ended_at + interval '3 H' as es_ended_at,
            e.slug
    from    api_clients.event_sessions as es
            join api_clients.events as e
                on e.id = es.event_id
    where   1=1
            and es.created_at >= '{date_from}'::date - interval '3 H'
            and es.created_at < '{date_to}'::date - interval '3 H'
            and es.ended_at is not null
            and e.br_parent_id is null
    ;""")
    all_confs_list.extend(cursor.fetchall())
    return all_confs_list, cursor, connection


@db_retry_decorator
def get_user_sessions_data(user_sessions_list: list, date_from: date, date_to: date, cursor=None, connection=None):
    """Выгружает пользовательские сессии по комнатам за период, начиная с **date_from** и до **date_to**

    :return user_sessions_list: [[us_created_at, us_ended_at, event_id, user_id, slug], ...]
    :type user_sessions_list: list[list]
    :param date_from: дата начала периода
    :type date_from: date
    :param date_to: дата конца периода
    :type date_to: date
    :param cursor: объект курсора подключения к БД
    :type cursor: cursor | None
    :param connection: объект подключения к БД
    :type connection: connection | None
    :return user_sessions_list: [[us_created_at, us_ended_at, event_id, user_id, slug], ...] (**list[list]**)
    :return cursor: объект курсора подключения к БД (**cursor** | **None**)
    :return connection: объект подключения к БД (**connection** | **None**)
    """
    cursor.execute(f"""
    SELECT  us.created_at + interval '3 H' as created_at,
            coalesce(us.ended_at, us.created_at + interval '1 min') + interval '3 H' as ended_at,
            us.event_id,
            us.user_id,
            e.slug
    FROM    api_clients.user_sessions as us
            JOIN api_clients.events as e
                on us.event_id = e.id
    where   1=1
            and (us.created_at, coalesce(us.ended_at, us.created_at + interval '1 min')) overlaps ('{date_from}'::date - interval '3 H', '{date_to}'::date - interval '3 H')
    ;""")
    user_sessions_list.extend(cursor.fetchall())
    return user_sessions_list, cursor, connection


def calculate_unique_users_qty(user_sessions_dic: dict, **es_dic):
    """Подсчет уникальных юзеров во время определенной ивент сессии (конференции).

    :param user_sessions_dic: сессии юзеров по комнатам {event_id: [[us_created_at, user_id], ...], ...}
    :type user_sessions_dic: dict
    :param es_dic: данные о конфе {'event_id': ID комнаты (str), 'event_session_created_at': начало конфы (datetime), 'event_session_ended_at': конец конфы (datetime)}
    :type es_dic: **dict
    :return unique_users_qty: кол-во уникальных user_id (**int**)
    :return users_dic: минимальное время захода в конфу для каждого user_id (**dict**)
    """
    event_id = es_dic['event_id']
    es_created_at = es_dic['event_session_created_at']
    es_ended_at = es_dic['event_session_ended_at']
    users_dic = {}
    unique_users_qty = 0
    user_sessions_list: list = user_sessions_dic.get(event_id, [])
    if user_sessions_list:
        user_sessions_list.sort(key=lambda x: x[0])
        for user_session in user_sessions_list:
            us_created_at, user_id = user_session
            if (es_created_at <= us_created_at) & (us_created_at <= es_ended_at):
                min_us_created_at = users_dic.get(user_id, us_created_at)
                users_dic[user_id] = min(min_us_created_at, us_created_at)
                unique_users_qty = len(users_dic)
            else:
                if unique_users_qty > 0:
                    break
    return unique_users_qty, users_dic


def merge_and_transform_data(db_confs: list, all_confs_list: list, user_sessions_list: list):
    """Возвращает сводную таблицу по дням и дименшенам. Собирает данные из **all_confs_list** и **user_sessions_list**.
    Проверяет актуальное состояние параметра комнаты с помощью **audit_event_param_dic**.
    Преобразует в таблицу для записи в БД, заполняя пустые частые сочетания дименшенов нулями

    :param db_confs: сводная таблица [[date_val, confs_qty], ...]
    :type db_confs: list[list]
    :param all_confs_list: [[es_id, event_id, es_created_at, es_ended_at, slug], ...]
    :type all_confs_list: list[list]
    :param user_sessions_list: [[us_created_at, us_ended_at, event_id, user_id, slug], ...]
    :type user_sessions_list: list[list]
    :return db_confs: сводная таблица [[date_val, confs_qty], ...] (**list[list]**)
    :return pd_conf_calls_dic: {slug: [[es_id, es_created_at, es_ended_at, {user_id: min_us_created_at, ...}], ...], ...} (**dict**)
    """
    user_sessions_dic = {}
    for row in user_sessions_list:
        us_created_at, us_ended_at, event_id, user_id, slug = row
        user_sessions_list: list = user_sessions_dic.get(event_id, [])
        user_sessions_list.append([us_created_at, user_id])
        user_sessions_dic[event_id] = user_sessions_list
    pd_conf_dic = {}
    pd_conf_calls_dic = {}
    for row in all_confs_list:
        es_id, event_id, es_created_at, es_ended_at, slug = row
        date_val = es_created_at.date()
        es_duration_sec = int((es_ended_at - es_created_at).total_seconds())
        es_params_dic = {'event_id': event_id, 'event_session_created_at': es_created_at, 'event_session_ended_at': es_ended_at}
        users_qty, users_dic = calculate_unique_users_qty(user_sessions_dic, **es_params_dic)
        slug_conf_list: list = pd_conf_calls_dic.get(slug, [])
        slug_conf_list.append([es_id, es_created_at, es_ended_at, users_dic])
        pd_conf_calls_dic[slug] = slug_conf_list
        if (users_qty >= 2) & (es_duration_sec >= 120):
            confs_set: set = pd_conf_dic.get(date_val, set())
            confs_set.add(es_id)
            pd_conf_dic[date_val] = confs_set
    for date_val in pd_conf_dic:
        confs_set: set = pd_conf_dic[date_val]
        confs_qty = len(confs_set)
        db_confs.append([date_val, confs_qty])
    return db_confs, pd_conf_calls_dic


@db_retry_decorator
def get_feedback_data(db_rating: list, date_from: date, date_to: date, cursor=None, connection=None):
    """Выгружает сумму и кол-во оценок на дату из **feed_feedback_schema.feedbacks** за период, начиная с **date_from** и до **date_to**

    :param db_rating: [[date_val, rating_sum, rating_qty], ...]
    :type db_rating: list[list]
    :param date_from: дата начала периода
    :type date_from: date
    :param date_to: дата конца периода
    :type date_to: date
    :param cursor: объект курсора подключения к БД
    :type cursor: cursor | None
    :param connection: объект подключения к БД
    :type connection: connection | None
    :return db_rating: [[date_val, rating_sum, rating_qty], ...] (**list[list]**)
    :return cursor: объект курсора подключения к БД (**cursor** | **None**)
    :return connection: объект подключения к БД (**connection** | **None**)
    """
    cursor.execute(f"""
    select  (created_at + interval '3 H')::date as date_val,
            sum(rating) as rating_sum,
            count(id) as rating_qty
    from    feed_feedback_schema.feedbacks
    where   1=1
            and created_at >= '{date_from}'::date - interval '3 H'
            and created_at < '{date_to}'::date - interval '3 H'
            and rating is not null
    group by
            date_val
    ;""")
    feedback_list = cursor.fetchall()
    if feedback_list:
        db_rating.extend(feedback_list)
    else:
        db_rating.append([date_from, 0, 0])
    return db_rating, cursor, connection


@db_retry_decorator
def get_reg_users_data(db_regs: list, date_from: date, date_to: date, cursor=None, connection=None):
    """Выгружает регистрации пользователей на дату из **api_clients.users** за период, начиная с **date_from** и до **date_to**

    :param db_regs: [[date_val, reg_users_qty], ...]
    :type db_regs: list[list]
    :param date_from: дата начала периода
    :type date_from: date
    :param date_to: дата конца периода
    :type date_to: date
    :param cursor: объект курсора подключения к БД
    :type cursor: cursor | None
    :param connection: объект подключения к БД
    :type connection: connection | None
    :return db_regs: [[date_val, reg_users_qty], ...] (**list[list]**)
    :return cursor: объект курсора подключения к БД (**cursor** | **None**)
    :return connection: объект подключения к БД (**connection** | **None**)
    """
    cursor.execute(f"""
    select  (created_at + interval '3 H')::date as date_val,
            count(id) as reg_users_qty
    from    api_clients.users
    where   1=1
            and created_at >= '{date_from}'::date - interval '3 H'
            and created_at < '{date_to}'::date - interval '3 H'
            and email not like 'guest_email_%'
            and email not like 'sip_user_email_%'
    group by
            date_val
    ;""")
    reg_users_list = cursor.fetchall()
    if reg_users_list:
        db_regs.extend(reg_users_list)
    else:
        db_regs.append([date_from, 0])
    return db_regs, cursor, connection


@db_retry_decorator
def get_active_users_qty(db_active_users: list, date_from: date, date_to: date, cursor=None, connection=None):
    """Выгружает текущее кол-во пользователей в статусе `active` из **api_clients.users** на дату **date_to**

    :return db_active_users: [[date_val, active_users_qty], ...]
    :type db_active_users: list[list]
    :param date_from: дата начала периода
    :type date_from: date
    :param date_to: дата конца периода
    :type date_to: date
    :param cursor: объект курсора подключения к БД
    :type cursor: cursor | None
    :param connection: объект подключения к БД
    :type connection: connection | None
    :return db_active_users: [[date_val, active_users_qty], ...] (**list[list]**)
    :return cursor: объект курсора подключения к БД (**cursor** | **None**)
    :return connection: объект подключения к БД (**connection** | **None**)
    """
    cursor.execute(f"""
    select  count(id) as reg_users_qty
    from    api_clients.users
    where   1=1
            and email not like 'guest_email_%'
            and email not like 'sip_user_email_%'
            and status = 'active'
            and deleted_at is null
    ;""")
    sql_request_result = cursor.fetchall()
    active_users_qty = sql_request_result[0][0] if sql_request_result else 0
    db_active_users.append([date_to - timedelta(days=1), active_users_qty])
    return db_active_users, cursor, connection


@db_retry_decorator
def get_calls_data(calls_list: list, date_from: date, date_to: date, cursor=None, connection=None):
    """Выгружает звонки на дату из **callserver.calls** за период, начиная с **date_from** и до **date_to**

    :param calls_list: пустой список для хранения звонков []
    :type calls_list: list
    :param date_from: дата начала периода
    :type date_from: date
    :param date_to: дата конца периода
    :type date_to: date
    :param cursor: объект курсора подключения к БД
    :type cursor: cursor | None
    :param connection: объект подключения к БД
    :type connection: connection | None
    :return calls_list: [[call_created_at, call_id, slug, caller_user_id, target_user_id, status, is_conference_call, group_call_id], ...] (**list[list]**)
    :return cursor: объект курсора подключения к БД (**cursor** | **None**)
    :return connection: объект подключения к БД (**connection** | **None**)
    """
    cursor.execute(f"""
    select  created_at + interval '3 hour' as call_created_at,
            id as call_id,
            slug,
            caller_user_id,
            target_user_id,
            status,
            is_conference_call,
            group_call_id
    from    callserver.calls
    where   1=1
            and created_at >= '{date_from}'::date - interval '3 H'
            and created_at < '{date_to}'::date - interval '3 H'
    ;""")
    calls_list.extend(cursor.fetchall())
    return calls_list, cursor, connection


def merge_calls_and_conference(db_calls: list, calls_list: list, pd_conf_calls_dic: dict, date_from: date):
    """Соединяет звонки **calls_list** и соответствующие им конференции из **pd_conf_calls_dic** на дату **date_from*.
    Агрегирует данные, добавляет в **db_calls** и возвращает его с переведенными на русский язык значениями

    :param db_calls: [[date_val, status, call_type, success_type, calls_qty], ...]
    :type db_calls: list[list]
    :param calls_list: [[call_created_at, call_id, slug, caller_user_id, target_user_id, status, is_conference_call, group_call_id], ...]
    :type calls_list: list[list]
    :param pd_conf_calls_dic: {slug: [[es_id, es_created_at, es_ended_at, {user_id: min_us_created_at, ...}], ...], ...}
    :type pd_conf_calls_dic: dict
    :param date_from: дата начала периода
    :type date_from: date
    :return db_calls: [[date_val, status, call_type, success_type, calls_qty], ...] (**list[list]**)
    """
    # Словари перевода на русский язык
    status_translations = {
        'succeeded': 'Успешно',
        'failed': 'Неуспешно',
        'cancelled': 'Отменено',
        'declined_by_caller': 'Отклонено звонящим',
        'not_answered': 'Без ответа',
        'busy': 'Занято',
        'declined_by_target': 'Отклонено принимающим'
    }
    call_type_translations = {
        '1vs1': 'Один на один',
        'group_call': 'Групповой звонок',
        'conference_call': 'Конференц-звонок'
    }
    success_type_translations = {
        'both': 'Оба участника',
        'only_caller': 'Только звонящий',
        'only_target': 'Только принимающий',
        'nobody': 'Никто'
    }

    calls_dic = {}
    if calls_list:
        for row in calls_list:
            call_created_at, call_id, slug, caller_user_id, target_user_id, status, is_conference_call, group_call_id = row
            call_is_added = False
            if not is_conference_call and not group_call_id:
                call_type = '1vs1'
            elif group_call_id:
                call_type = 'group_call'
            elif is_conference_call:
                call_type = 'conference_call'
            else:
                call_type = None
            if status == 'succeeded':
                slug_conf_list: list = pd_conf_calls_dic.get(slug, [])
                slug_conf_list.sort(key=lambda x: x[1])
                if slug_conf_list:
                    for conf_data in slug_conf_list:
                        es_id, es_created_at, es_ended_at, users_dic = conf_data
                        if caller_user_id in users_dic and target_user_id in users_dic:
                            success_type = 'both'
                        elif caller_user_id in users_dic:
                            success_type = 'only_caller'
                        elif target_user_id in users_dic:
                            success_type = 'only_target'
                        else:
                            success_type = 'nobody'
                        if call_type == 'conference_call':
                            if (es_created_at <= call_created_at) & (call_created_at <= es_ended_at):
                                success_type = None
                                calls_set: set = calls_dic.get((date_from, status, call_type, success_type), set())
                                calls_set.add(call_id)
                                calls_dic[(date_from, status, call_type, success_type)] = calls_set
                                call_is_added = True
                        elif call_type == 'group_call':
                            success_type = None
                            if es_created_at > call_created_at:
                                call_conf_delta = es_created_at - call_created_at
                            else:
                                call_conf_delta = call_created_at - es_created_at
                            if call_conf_delta <= timedelta(seconds=60):
                                calls_set: set = calls_dic.get((date_from, status, call_type, success_type), set())
                                calls_set.add(call_id)
                                calls_dic[(date_from, status, call_type, success_type)] = calls_set
                                call_is_added = True
                        else:
                            calls_set: set = calls_dic.get((date_from, status, call_type, success_type), set())
                            calls_set.add(call_id)
                            calls_dic[(date_from, status, call_type, success_type)] = calls_set
                            call_is_added = True
                else:
                    success_type = 'nobody'
                    calls_set: set = calls_dic.get((date_from, status, call_type, success_type), set())
                    calls_set.add(call_id)
                    calls_dic[(date_from, status, call_type, success_type)] = calls_set
                    call_is_added = True
            else:
                success_type = None
                calls_set: set = calls_dic.get((date_from, status, call_type, success_type), set())
                calls_set.add(call_id)
                calls_dic[(date_from, status, call_type, success_type)] = calls_set
                call_is_added = True
            if not call_is_added:
                if success_type:
                    success_type = 'nobody'
                calls_set: set = calls_dic.get((date_from, status, call_type, success_type), set())
                calls_set.add(call_id)
                calls_dic[(date_from, status, call_type, success_type)] = calls_set
                call_is_added = True
    if calls_dic:
        for kkey_tuple in calls_dic:
            date_val, status, call_type, success_type = kkey_tuple
            calls_set: set = calls_dic[kkey_tuple]
            calls_qty = len(calls_set)
            # Переводим значения на русский
            status = status_translations.get(status, status)
            call_type = call_type_translations.get(call_type, call_type)
            success_type = success_type_translations.get(success_type, success_type)
            db_calls.append([date_val, status, call_type, success_type, calls_qty])
    return db_calls


def create_structure(wb, min_date, max_date, sheets_data):
    """Создает оглавление, все листы и ссылки из оглавления на листы"""
    ws_title = wb.active
    ws_title.title = 'ОГЛАВЛЕНИЕ'
    ws_title.column_dimensions["A"].width = 77
    ws_title['A1'] = 'Отчет с {} по {} (таймзона UTC+3)'.format(min_date, max_date)
    ws_title['A1'].font = Font(name='Calibri', size=20, bold=True)
    ws_title['A4'] = 'Оглавление'
    ws_title['A4'].font = Font(name='Calibri', size=20, bold=True)
    for x in range(len(sheets_data)):
        data, sheet_name, toc_name, table_title, period_key = sheets_data[x]
        wb.create_sheet(title="{}".format(sheet_name))
        link = "#'{}'!A1".format(sheet_name)
        ws_title['A{}'.format(str(4+1+x))] = '=HYPERLINK("{}", "{}")'.format(link, toc_name)
        ws_title['A{}'.format(str(4+1+x))].font = Font(name='Calibri', size=20, color='0000FF', underline='single')
    return wb


def create_data_for_sheet(data, min_date, max_date, period_type, calc_rating=False, calc_max=False):
    """Группирует данные в зависимости от типа периода и типа данных"""
    data_for_sheet = []
    periods = set()
    while min_date <= max_date:
        if period_type == 'pd':
            periods.add(min_date)
        elif period_type == 'pw':
            week = str(min_date.isocalendar()[1])
            week = '0'+week if len(week) == 1 else week
            periods.add(str(min_date.isocalendar()[0]) + '-' + week)
        elif period_type == 'pm':
            month = str(min_date.month)
            month = '0' + month if len(month) == 1 else month
            periods.add(str(min_date.year) + '-' + month)
        min_date = min_date + timedelta(days=1)
    periods = list(periods)
    periods.sort()
    data_dic = {}
    for x in data:
        if period_type == 'pd':
            period_name = x[0]
        elif period_type == 'pw':
            week = str(x[0].isocalendar()[1])
            week = '0'+week if len(week) == 1 else week
            period_name = (str(x[0].isocalendar()[0]) + '-' + week)
        elif period_type == 'pm':
            month = str(x[0].month)
            month = '0' + month if len(month) == 1 else month
            period_name = (str(x[0].year) + '-' + month)
        if calc_rating:
            period_value = data_dic.get(period_name, [0, 0])
        else:
            period_value = data_dic.get(period_name, 0)
        if calc_rating:
            period_value[0] += x[1]
            period_value[1] += x[2]
        else:
            if calc_max:
                if x[1] > period_value:
                    period_value = x[1]
            else:
                period_value += x[1]
        data_dic[period_name] = period_value
    if calc_rating:
        for x in periods:
            try:
                val = round(data_dic.get(x, [0, 0])[0] / data_dic.get(x, [0, 0])[1], 2)
            except:
                val = 0
            data_for_sheet.append([x, val])
    else:
        [data_for_sheet.append([x, data_dic.get(x, 0)]) for x in periods]
    return data_for_sheet


def draw_sheet(wb, ws_data, min_date, max_date):
    """Рисует лист в книге с данными и графиками"""
    data, sheet_name, toc_name, table_title, period_key = ws_data
    int_number_format = '# ##0'
    float_number_format = '0.00'
    number_format = int_number_format
    period_name_dic = {'pd': 'Дата', 'pw': 'Неделя', 'pm': 'Месяц'}
    period_name = period_name_dic.get(period_key, '')
    ws = wb[sheet_name]
    if table_title == 'db_calls':
        ws['A1'] = '=HYPERLINK("{}", "{}")'.format("#'ОГЛАВЛЕНИЕ'!A1", "Вернуться к Оглавлению")
        ws['A1'].font = Font(name='Calibri', size=20, color='0000FF', underline='single')
        column_names = ['Дата', 'Статус', 'Тип звонка', 'Присутствие участников (для 1vs1)', 'Количество звонков']
        column_widths = [len(x) for x in column_names]
        columns_qty = len(column_names)
        sr, sc = (3, 1)
        all_org_data = []
        all_org_data.append(column_names)
        all_org_data.extend(data)
        rows_qty = len(all_org_data)
        for i, row in enumerate(all_org_data, start=sr):
            for j, val in enumerate(row, start=sc):
                ws_cell = ws.cell(row=i, column=j, value=val)
                if isinstance(val, (date, datetime)):
                    ws_cell.number_format = 'DD.MM.YYYY'
                    if column_widths[j-sc] < 11:
                        column_widths[j-sc] = 11
                elif isinstance(val, float):
                    ws_cell.number_format = float_number_format
                elif isinstance(val, int):
                    ws_cell.number_format = int_number_format
                if len(str(val)) > column_widths[j-sc]:
                    column_widths[j-sc] = len(str(val))
        for j, column_width in enumerate(column_widths, start=sc):
            adjusted_width = column_width * 0.8 + 6
            ws.column_dimensions[get_column_letter(j)].width = adjusted_width
        ref_all_data = f"{get_column_letter(sc)}{sr}:{get_column_letter(sc+columns_qty-1)}{sr+rows_qty-1}"
        filters = ws.auto_filter
        filters.ref = ref_all_data
        explanation_row = sr + rows_qty + 1
        ws[f'A{explanation_row}'] = 'Пояснения:'
        ws[f'A{explanation_row}'].font = Font(bold=True)
        explanations = [
            ('Тип звонка:', 'Один на один - звонок между двумя пользователями'),
            ('', 'Групповой звонок - звонок с участием более двух пользователей'),
            ('', 'Конференц-звонок - звонок в рамках конференции'),
            ('Статус:', 'Успешно - звонок был установлен'),
            ('', 'Неуспешно - звонок не был установлен'),
            ('', 'Отменено - звонок был отменен'),
            ('', 'Отклонено звонящим - звонящий отклонил звонок'),
            ('', 'Без ответа - принимающий не ответил на звонок'),
            ('', 'Занято - принимающий был занят'),
            ('', 'Отклонено принимающим - принимающий отклонил звонок'),
            ('Присутствие участников (для 1vs1):', 'Оба участника - и звонящий, и принимающий были в звонке'),
            ('', 'Только звонящий - только звонящий был в звонке'),
            ('', 'Только принимающий - только принимающий был в звонке'),
            ('', 'Никто - ни один из участников не был в звонке')
        ]
        for idx, (title, text) in enumerate(explanations):
            ws[f'A{explanation_row + idx + 1}'] = title
            ws[f'B{explanation_row + idx + 1}'] = text
    elif table_title:
        if 'Средняя оценка' in sheet_name:
            number_format = float_number_format
            data = create_data_for_sheet(data, min_date, max_date, period_key, calc_rating=True)
        elif 'Макс пользователей' in sheet_name or 'аккаунты' in sheet_name:
            data = create_data_for_sheet(data, min_date, max_date, period_key, calc_max=True)
        else:
            data = create_data_for_sheet(data, min_date, max_date, period_key)
        ws.column_dimensions["A"].width = 20
        ws.column_dimensions["B"].width = 53
        ws['A1'] = '=HYPERLINK("{}", "{}")'.format("#'ОГЛАВЛЕНИЕ'!A1", "Вернуться к Оглавлению")
        ws['A1'].font = Font(name='Calibri', size=20, color='0000FF', underline='single')
        title_list = re.split(r'\b(по|в|за)\b', toc_name)
        value_name = title_list[0].strip()
        data_row_start = 50
        ws['A{}'.format(str(data_row_start))] = period_name
        ws['B{}'.format(str(data_row_start))] = value_name
        for x in range(len(data)):
            ws['A{}'.format(str(data_row_start + 1 + x))] = data[x][0]
            ws['B{}'.format(str(data_row_start + 1 + x))] = data[x][1]
            ws['B{}'.format(str(data_row_start + 1 + x))].number_format = number_format
            data_row_end = data_row_start + 1 + x
        data_row_start += 1
        c1 = LineChart()
        if 'Макс пользователей' in sheet_name:
            title_val = toc_name
        elif 'Макс конференций' in sheet_name:
            title_val = toc_name + ' (длительность больше 2 минут и больше 1 участника)'
        elif 'Средняя оценка' in sheet_name:
            title_val = toc_name
        elif 'Новых пользователей' in sheet_name:
            title_val = toc_name
        else:
            title_val = toc_name
        c1.title = title_val
        c1.style = 2
        c1.y_axis.title = value_name
        c1.x_axis.title = period_name
        c1.x_axis.delete = False
        c1.y_axis.delete = False
        c1.width = 60
        c1.height = 20
        values = Reference(ws, min_col=2, min_row=data_row_start-1, max_row=data_row_end)
        c1.add_data(values, titles_from_data=True)
        dates = Reference(ws, min_col=1, min_row=data_row_start, max_row=data_row_end)
        c1.set_categories(dates)
        s1 = c1.series[0]
        s1.marker.symbol = "diamond"
        c1.dataLabels = DataLabelList(dLblPos='t', showVal=True, showSerName=False, showLegendKey=False, showCatName=False)
        if period_key == 'pd':
            c1.dataLabels.txPr = RichText(p=[Paragraph(pPr=ParagraphProperties(defRPr=CharacterProperties(sz=900)),
                                                    endParaRPr=CharacterProperties(sz=900))])
        else:
            c1.dataLabels.txPr = RichText(p=[Paragraph(pPr=ParagraphProperties(defRPr=CharacterProperties(sz=1400)),
                                                    endParaRPr=CharacterProperties(sz=1400))])
        c1.x_axis.title.tx.rich.p[0].r[0].rPr = CharacterProperties(b=True)
        c1.legend = None
        ws.add_chart(c1, "A2")
    else:
        ws.column_dimensions["A"].width = 40
        ws.column_dimensions["B"].width = 9
        ws['A1'] = '=HYPERLINK("{}", "{}")'.format("#'ОГЛАВЛЕНИЕ'!A1", "Вернуться к Оглавлению")
        ws['A1'].font = Font(name='Calibri', size=20, color='0000FF', underline='single')
        data_row_start = 3
        ws[f'A{data_row_start}'] = toc_name
        ws[f'B{data_row_start}'] = data[0][1]
        ws[f'B{data_row_start}'].number_format = number_format
    return wb


def create_xlsx(db_confs, db_rating, db_regs, db_maxload, db_active_users, db_calls, dest_dit: str, file_name: str):
    """Создает XLSX отчет"""
    dates_set = set()
    [dates_set.add(x[0]) for x in db_confs if db_confs]
    [dates_set.add(x[0]) for x in db_regs if db_regs]
    [dates_set.add(x[0]) for x in db_rating if db_rating]
    [dates_set.add(x[0]) for x in db_active_users if db_active_users]
    [dates_set.add(x[0]) for x in db_maxload if db_maxload]
    min_date = min(dates_set)
    max_date = max(dates_set)
    sheets_data = [
        (db_active_users, 'Пользователи в статусе active', 'Пользователи в статусе active', None, None),
        (db_maxload, 'Макс пользователей в день', 'Максимальное количество одновременных пользователей в день', True, 'pd'),
        (db_maxload, 'Макс пользователей в неделю', 'Максимальное количество одновременных пользователей в неделю', True, 'pw'),
        (db_maxload, 'Макс пользователей в месяц', 'Максимальное количество одновременных пользователей в месяц', True, 'pm'),
        (db_confs, 'Макс конференций в день', 'Количество конференций в день', True, 'pd'),
        (db_confs, 'Макс конференций в неделю', 'Количество конференций в неделю', True, 'pw'),
        (db_confs, 'Макс конференций в месяц', 'Количество конференций в месяц', True, 'pm'),
        (db_rating, 'Средняя оценка в день', 'Средняя оценка за день', True, 'pd'),
        (db_rating, 'Средняя оценка в неделю', 'Средняя оценка за неделю', True, 'pw'),
        (db_rating, 'Средняя оценка в месяц', 'Средняя оценка за месяц', True, 'pm'),
        (db_regs, 'Новых пользователей в день', 'Количество новых пользователей в день', True, 'pd'),
        (db_regs, 'Новых пользователей в неделю', 'Количество новых пользователей в неделю', True, 'pw'),
        (db_regs, 'Новых пользователей в месяц', 'Количество новых пользователей в месяц', True, 'pm'),
        (db_calls, 'Звонки', 'Звонки', 'db_calls', None),
    ]
    wb = Workbook()
    if not os.path.exists(dest_dit):
        os.makedirs(dest_dit)
    dest_filename = os.path.join(dest_dit, file_name)
    wb = create_structure(wb, min_date, max_date, sheets_data)
    for ws_data in sheets_data:
        wb = draw_sheet(wb, ws_data, min_date, max_date)
    for i in range(2):
        try:
            wb.save(filename=dest_filename)
            print(f'Файл-отчет сохранен в {dest_filename}')
            break
        except Exception as e:
            print(f'{e}')
            input('Закройте файл и нажмите Enter: ')
    wb.close()
    return dest_filename


def get_all_db_data(date_from: date, date_to: date):
    """Получает все данные для отчета за период, начиная с **date_from** и до **date_to**

    :param date_from: дата начала выгрузки
    :type date_from: date
    :param date_to: дата конца выгрузки
    :type date_to: date
    :return db_confs: [[date_val, confs_qty], ...] (**list[list]**)
    :return db_maxload: [[date_val, max_val], ...] (**list[list]**)
    :return db_rating: [[date_val, rating_sum, rating_qty], ...] (**list[list]**)
    :return db_regs: [[date_val, ...], ...] (**list[list]**)
    :return db_active_users: [[date_val, ...], ...] (**list[list]**)
    :return db_calls: [[date_val, status, call_type, success_type, calls_qty], ...] (**list[list]**)
    """
    db_confs = []
    db_maxload = []
    db_rating = []
    db_regs = []
    db_active_users = []
    db_calls = []
    ddate_to = date_to
    connection = pg_replica_connection()
    cursor = connection.cursor()
    while date_from < ddate_to:
        date_to = date_from + timedelta(days=1)
        print(f"{date_from}")
        all_confs_list = []
        user_sessions_list = []
        all_confs_list, cursor, connection = get_all_confs_data(all_confs_list, date_from, date_to, cursor, connection)
        user_sessions_list, cursor, connection = get_user_sessions_data(user_sessions_list, date_from, date_to, cursor, connection)
        if all_confs_list and user_sessions_list:
            db_confs, pd_conf_calls_dic = merge_and_transform_data(db_confs, all_confs_list, user_sessions_list)
        else:
            db_confs.append([date_from, 0])
            pd_conf_calls_dic = {}
        if user_sessions_list:
            grouped_us_dic, min_dt, max_dt = process_db(user_sessions_list, date_from, date_to)
            ts_sec_array, ts_sec_idx_dic = divide_period_per_second(min_dt, max_dt)
            db_maxload = calculate_max_concurrent_for_db(db_maxload, grouped_us_dic, ts_sec_array, ts_sec_idx_dic)
        else:
            db_maxload.append([date_from, 0])
        db_rating, cursor, connection = get_feedback_data(db_rating, date_from, date_to, cursor, connection)
        db_regs, cursor, connection = get_reg_users_data(db_regs, date_from, date_to, cursor, connection)
        calls_list = []
        calls_list, cursor, connection = get_calls_data(calls_list, date_from, date_to, cursor, connection)
        db_calls = merge_calls_and_conference(db_calls, calls_list, pd_conf_calls_dic, date_from)
        date_from = date_to
    db_active_users, cursor, connection = get_active_users_qty(db_active_users, None, ddate_to, cursor, connection)
    cursor.close()
    connection.close()
    return db_confs, db_maxload, db_rating, db_regs, db_active_users, db_calls


def execute():
    """Запускает скрипт"""
    date_from = date.today() - timedelta(days=DEPTH_DAYS_QTY)
    date_to = date.today() + timedelta(days=1)
    dest_dir = f'{os.sep}'.join(FULL_FILENAME_XLSX.split(os.sep)[:-1])
    file_name = FULL_FILENAME_XLSX.split(os.sep)[-1]
    print(f"Данные за период от: {date_from}\tдо: {date_to-timedelta(days=1)} (включительно)")
    print('Начинается выгрузка по дням...')
    db_confs, db_maxload, db_rating, db_regs, db_active_users, db_calls = get_all_db_data(date_from, date_to)
    print('Done!')
    create_xlsx(db_confs, db_rating, db_regs, db_maxload, db_active_users, db_calls, dest_dir, file_name)


if __name__ == "__main__":
    dts = datetime.now()
    print(f'start at:\t{dts.strftime("%H:%M:%S %Y-%m-%d")}')
    execute()
    print(f"Весь скрипт: {datetime.now()-dts}".split('.')[0])